#pragma once

#include <QDialog>
#include <QVariant>

namespace Ui {
class PlatformDialog;
}

class PlatformDialog : public QDialog
{
    Q_OBJECT

public:
    explicit PlatformDialog(QWidget *parent = 0);
    ~PlatformDialog();

    void addPlatform(QString name);
    bool useCpu();
    bool saveTemps();
    bool wError();
    bool fastMath();
    QString currentPlatform();

private:
    Ui::PlatformDialog *ui;
};

